@@ -1,5 +1,5 @@
 /************************************************************************************
-   Copyright (C) 2020,2022 MariaDB Corporation AB
+   Copyright (C) 2020,2025 MariaDB Corporation plc
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -131,6 +131,10 @@ SQLRETURN MA_SQLAllocHandle(SQLSMALLINT HandleType,
     }
     return SQL_ERROR;
   }
+  catch (SQLException &e)
+  {
+    ret= MADB_FromException(*Error, e);
+  }
   return ret;
 }
 /* }}} */
@@ -146,6 +150,8 @@ SQLRETURN MA_SQLBindCol(SQLHSTMT StatementHandle,
   MADB_Stmt* Stmt= (MADB_Stmt*)StatementHandle;
   SQLRETURN ret;
 
+  RESET_CANCELED(Stmt);
+
   MDBUG_C_ENTER(Stmt->Connection, "SQLBindCol");
   MDBUG_C_DUMP(Stmt->Connection, Stmt, 0x);
   MDBUG_C_DUMP(Stmt->Connection, ColumnNumber, u);
@@ -161,7 +167,10 @@ SQLRETURN MA_SQLBindCol(SQLHSTMT StatementHandle,
   {
     ret= MADB_SetError(&Stmt->Error, MADB_ERR_HY001, NULL, 0);
   }
-
+  catch (SQLException &e)
+  {
+    ret= MADB_FromException(Stmt->Error, e);
+  }
   MDBUG_C_RETURN(Stmt->Connection, ret, &Stmt->Error);
 }
 /* }}} */
@@ -182,9 +191,7 @@ SQLRETURN MA_SQLBindParameter(SQLHSTMT StatementHandle,
   SQLRETURN ret;
 
   MADB_CLEAR_ERROR(&((MADB_Stmt*)StatementHandle)->Error);
-
-  if (!Stmt)
-    return SQL_INVALID_HANDLE;
+  RESET_CANCELED(Stmt);
 
   MDBUG_C_ENTER(Stmt->Connection, "SQLBindParameter");
   MDBUG_C_DUMP(Stmt->Connection, Stmt, 0x);
@@ -207,7 +214,10 @@ SQLRETURN MA_SQLBindParameter(SQLHSTMT StatementHandle,
   {
     ret= MADB_SetError(&Stmt->Error, MADB_ERR_HY001, NULL, 0);
   }
-
+  catch (SQLException &e)
+  {
+    ret= MADB_FromException(Stmt->Error, e);
+  }
   MDBUG_C_RETURN(Stmt->Connection, ret, &Stmt->Error);
 }
 /* }}} */
@@ -262,62 +272,97 @@ SQLRETURN MA_SQLBulkOperations(SQLHSTMT StatementHandle,
 SQLRETURN MA_SQLCancel(SQLHSTMT StatementHandle)
 {
   MADB_Stmt *Stmt= (MADB_Stmt *)StatementHandle;
-  SQLRETURN ret= SQL_ERROR;
+  SQLRETURN ret= SQL_SUCCESS;
 
   if (!Stmt)
     return SQL_INVALID_HANDLE;
 
+  // Technically, if the application does not do good syncronization of stmt use here, we can have
+  // Stmt already freed at this point, and crash right away on reading the mutex in Stmt
+  std::unique_lock<std::mutex> localScopeLock(globalLock);
+  if (CheckDeletedStmt(Stmt) != NULL)
+  {
+    return ret;// i.e. SQL_SUCCESS. Stmt has been deleted, nothing to cancel already
+  }
+  if (!Stmt->CancelDropSwitch.try_lock())
+  {
+    // This is not clear what is right here - success or error
+    return SQL_SUCCESS;
+  }
+  // We can release the lock. SQL_DROP will get it, but stop at the Stmt->CancelDropSwitch
+  // that we already own
+  localScopeLock.unlock();
   MADB_CLEAR_ERROR(&Stmt->Error);
 
   MDBUG_C_ENTER(Stmt->Connection, "SQLCancel");
   MDBUG_C_DUMP(Stmt->Connection, Stmt, 0x);
 
   auto& lock= Stmt->Connection->guard->getLock();
-  
-  if (lock.try_lock())
-  {
-    lock.unlock();
-    try
-    {
-      ret= Stmt->Methods->StmtFree(Stmt, SQL_CLOSE);
-    }
-    catch (std::bad_alloc &/*e*/)
-    {
-      ret= MADB_SetError(&Stmt->Error, MADB_ERR_HY001, NULL, 0);
-    }
-    catch (SQLException &e)
-    {
-      ret= MADB_FromException(Stmt->Error, e);
-    }
 
-    MDBUG_C_RETURN(Stmt->Connection, ret, &Stmt->Error);
+  /* In ODBC 2.x, if an application calls SQLCancel when no processing is being done on the statement,
+     SQLCancel has the same effect as SQLFreeStmt with the SQL_CLOSE option */
+  if (Stmt->Connection->Environment->OdbcVersion == SQL_OV_ODBC2)
+  {
+    if (lock.try_lock())
+    {
+      try
+      {
+        MADB_CloseCursor(Stmt);
+      }
+      catch (std::bad_alloc &/*e*/)
+      {
+        ret= MADB_SetError(&Stmt->Error, MADB_ERR_HY001, NULL, 0);
+      }
+      catch (SQLException &e)
+      {
+        ret= MADB_FromException(Stmt->Error, e);
+      }
+      lock.unlock();
+    }
+    else
+    {
+      ret= MADB_KillAtServer(Stmt->Connection, &Stmt->Error);
+    }
   }
   else
   {
-    // There is nothing, that can throw here
-    MYSQL *MariaDb;
-    
-    char StmtStr[30];
-
-    if (!(MariaDb= mysql_init(NULL)))
+    /* SQLCancel can cancel the following types of processing on a statement:
+       - A function running asynchronously on the statement -- those we do not have
+       - A function on a statement that needs data -- so first check this and reset it if needed
+       - A function running on the statement on another thread -- or killing process on server otherwise */
+    if (Stmt->PutParam > MADB_NO_DATA_NEEDED && !DAE_DONE(Stmt))
     {
-      ret= SQL_ERROR;
-      goto end;
-    }
-    if (!SQL_SUCCEEDED(Stmt->Connection->CoreConnect(MariaDb, Stmt->Connection->Dsn, &Stmt->Error)))
+      RESET_DAE_STATUS(Stmt);
+    } // Else we are canceling function running in other thread.
+    else if (lock.try_lock())
     {
-      mysql_close(MariaDb);
-      goto end;
+      Stmt->canceled= true;
+      try
+      {
+        /* "If a SQL statement is being executed when SQLCancel is called on another thread to cancel the
+          statement execution, it is possible for the execution to succeed and return SQL_SUCCESS while the
+          cancel is also successful. In this case, the Driver Manager assumes that the cursor opened by the
+          statement execution is closed by the cancel, so the application will not be able to use the cursor."
+       */
+        // It's not clear whether we should really close it here, though
+        MADB_CloseCursor(Stmt);
+      }
+      catch (std::bad_alloc &/*e*/)
+      {
+        ret= MADB_SetError(&Stmt->Error, MADB_ERR_HY001, NULL, 0);
+      }
+      catch (SQLException &e)
+      {
+        ret= MADB_FromException(Stmt->Error, e);
+      }
+      lock.unlock();
     }
-    
-    unsigned long len= static_cast<unsigned long>(_snprintf(StmtStr, 30, "KILL QUERY %ld", mysql_thread_id(Stmt->Connection->mariadb)));
-    if (!mysql_real_query(MariaDb, StmtStr, len))
+    else
     {
-      ret= SQL_SUCCESS;
+      ret= MADB_KillAtServer(Stmt->Connection, &Stmt->Error);
     }
-    mysql_close(MariaDb);
   }
-end:
+  Stmt->CancelDropSwitch.unlock();
   MDBUG_C_RETURN(Stmt->Connection, ret, &Stmt->Error);
 }
 /* }}} */
@@ -325,8 +370,14 @@ end:
 /* {{{ MA_SQLCancelDbc */
 SQLRETURN MA_SQLCancelDbc(SQLHANDLE Handle)
 {
-  MADB_Stmt Stmt(static_cast<MADB_Dbc*>(Handle));
-  return MA_SQLCancel((SQLHSTMT)&Stmt);
+  MADB_Dbc* Dbc(static_cast<MADB_Dbc*>(Handle));
+  auto& lock= Dbc->guard->getLock();
+
+  if (!lock.try_lock())
+  {
+    return MADB_KillAtServer(Dbc, &Dbc->Error);
+  }
+  return SQL_SUCCESS;
 }
 /* }}} */
 
@@ -337,6 +388,7 @@ SQLRETURN MA_SQLCloseCursor(SQLHSTMT StatementHandle)
   SQLRETURN ret;
 
   MADB_CLEAR_ERROR(&Stmt->Error);
+  RESET_CANCELED(Stmt);
 
   MDBUG_C_ENTER(Stmt->Connection, "SQLCloseCursor");
   MDBUG_C_DUMP(Stmt->Connection, StatementHandle, 0x);
@@ -384,6 +436,7 @@ SQLRETURN MA_SQLColAttribute(SQLHSTMT StatementHandle,
   SQLRETURN ret;
 
   MADB_CLEAR_ERROR(&Stmt->Error);
+  RESET_CANCELED(Stmt);
 
   MDBUG_C_ENTER(Stmt->Connection, "SQLColAttribute");
   MDBUG_C_DUMP(Stmt->Connection, StatementHandle, 0x);
@@ -737,6 +790,7 @@ SQLRETURN MA_SQLDescribeCol(SQLHSTMT StatementHandle,
   SQLRETURN ret;
 
   MADB_CLEAR_ERROR(&Stmt->Error);
+  RESET_CANCELED(Stmt);
 
   MDBUG_C_ENTER(Stmt->Connection, "SQLDescribeCol");
   MDBUG_C_DUMP(Stmt->Connection, Stmt, 0x);
@@ -1086,7 +1140,7 @@ SQLRETURN MA_SQLExecute(SQLHSTMT StatementHandle)
   
   try
   {
-    return Stmt->Methods->Execute(Stmt, FALSE);
+    return Stmt->Methods->Execute(Stmt, false);
   }
   catch (SQLException &e)
   {
@@ -1166,6 +1220,7 @@ SQLRETURN MA_SQLFetch(SQLHSTMT StatementHandle)
 {
   MADB_Stmt* Stmt= (MADB_Stmt*)StatementHandle;
 
+  RESET_CANCELED(Stmt);
   MDBUG_C_ENTER(Stmt->Connection, "SQLFetch");
 
   /* SQLFetch is equivalent of SQLFetchScroll(SQL_FETCH_NEXT), 3rd parameter is ignored for SQL_FETCH_NEXT */
@@ -1197,6 +1252,7 @@ SQLRETURN MA_SQLFetchScroll(SQLHSTMT StatementHandle,
 {
   MADB_Stmt* Stmt= (MADB_Stmt*)StatementHandle;
 
+  RESET_CANCELED(Stmt);
   MDBUG_C_ENTER(Stmt->Connection, "SQLFetchScroll");
   MDBUG_C_DUMP(Stmt->Connection, FetchOrientation, d);
 
@@ -1299,6 +1355,7 @@ SQLRETURN MA_SQLFreeStmt(SQLHSTMT StatementHandle,
 {
   MADB_Stmt* Stmt= (MADB_Stmt*)StatementHandle;
 
+  RESET_CANCELED(Stmt);
   MDBUG_C_ENTER(((MADB_Stmt*)StatementHandle)->Connection, "SQLFreeStmt");
   MDBUG_C_DUMP(Stmt->Connection, StatementHandle, 0x);
   MDBUG_C_DUMP(Stmt->Connection, Option, d);
@@ -1475,6 +1532,8 @@ SQLRETURN MA_SQLGetCursorName(
   int isWstr)
 {
   MADB_Stmt* Stmt= (MADB_Stmt*)StatementHandle;
+
+  RESET_CANCELED(Stmt);
   try
   {
     return Stmt->Methods->GetCursorName(Stmt, CursorName, BufferLength, NameLengthPtr, isWstr);
@@ -1483,6 +1542,10 @@ SQLRETURN MA_SQLGetCursorName(
   {
     return MADB_SetError(&Stmt->Error, MADB_ERR_HY001, NULL, 0);
   }
+  catch (SQLException &e)
+  {
+    return MADB_FromException(Stmt->Error, e);
+  }
   return SQL_SUCCESS;
 }
 /* }}} */
@@ -1498,19 +1561,26 @@ SQLRETURN MA_SQLGetData(SQLHSTMT StatementHandle,
   MADB_Stmt* Stmt= (MADB_Stmt*)StatementHandle;
   unsigned int i;
   MADB_DescRecord* IrdRec;
+  uint32_t columnCount= 0;
 
+  RESET_CANCELED(Stmt);
   /* In case we don't have DM(it check for that) */
   if (TargetValuePtr == NULL)
   {
     return MADB_SetError(&Stmt->Error, MADB_ERR_HY009, NULL, 0);
   }
-
   /* Bookmark */
   if (Col_or_Param_Num == 0)
   {
     return MADB_GetBookmark(Stmt, TargetType, TargetValuePtr, BufferLength, StrLen_or_IndPtr);
   }
-
+  /* This parameter validation has to be done before using it as array index
+   * To be on the safer side also checking if metadata is set, and if not - also treating it as index > max 
+   */
+  if (!Stmt->metadata || Col_or_Param_Num > (columnCount= Stmt->metadata->getColumnCount()))
+  {
+    return MADB_SetError(&Stmt->Error, MADB_ERR_07009, NULL, 0);
+  }
   /* We don't need this to be checked in case of "internal" use of the GetData, i.e. for internal needs we should always get the data */
   if (Stmt->CharOffset[Col_or_Param_Num - 1] > 0
     && Stmt->CharOffset[Col_or_Param_Num - 1] >= Stmt->Lengths[Col_or_Param_Num - 1])
@@ -1524,7 +1594,7 @@ SQLRETURN MA_SQLGetData(SQLHSTMT StatementHandle,
   }
 
   /* reset offsets for other columns. Doing that here since "internal" calls should not do that */
-  for (i= 0; i < Stmt->metadata->getColumnCount(); i++)
+  for (i= 0; i < columnCount; i++)
   {
     if (i != Col_or_Param_Num - 1)
     {
@@ -1743,6 +1813,8 @@ SQLRETURN MA_SQLGetStmtAttr(SQLHSTMT StatementHandle,
     SQLINTEGER *StringLengthPtr)
 {
   MADB_Stmt *Stmt= (MADB_Stmt *)StatementHandle;
+
+  RESET_CANCELED(Stmt);
   // GetAttr method will figure out if this is wide or ansi string
   try
   {
@@ -1801,6 +1873,10 @@ SQLRETURN MA_SQLGetFunctions(SQLHDBC ConnectionHandle,
   {
     ret= MADB_SetError(&Dbc->Error, MADB_ERR_HY001, NULL, 0);
   }
+  catch (SQLException &e)
+  {
+    ret= MADB_FromException(Dbc->Error, e);
+  }
   MDBUG_C_RETURN(Dbc, ret, &Dbc->Error);
 }
 /* }}} */
@@ -1896,6 +1972,7 @@ SQLRETURN MA_SQLNumParams(SQLHSTMT StatementHandle,
   SQLSMALLINT* ParameterCountPtr)
 {
   MADB_Stmt* Stmt= (MADB_Stmt*)StatementHandle;
+  RESET_CANCELED(Stmt);
   return Stmt->Methods->ParamCount(Stmt, ParameterCountPtr);
 }
 /* }}} */
@@ -1905,6 +1982,7 @@ SQLRETURN MA_SQLNumResultCols(SQLHSTMT StatementHandle,
   SQLSMALLINT* ColumnCountPtr)
 {
   MADB_Stmt* Stmt= (MADB_Stmt*)StatementHandle;
+  RESET_CANCELED(Stmt);
   return Stmt->Methods->ColumnCount(Stmt, ColumnCountPtr);
 }
 /* }}} */
@@ -1914,6 +1992,7 @@ SQLRETURN MA_SQLParamData(SQLHSTMT StatementHandle,
   SQLPOINTER* ValuePtrPtr)
 {
   MADB_Stmt* Stmt= (MADB_Stmt*)StatementHandle;
+  RESET_CANCELED(Stmt);
   try
   {
     return Stmt->Methods->ParamData(Stmt, ValuePtrPtr);
@@ -1936,6 +2015,7 @@ SQLRETURN MA_SQLPrepare(SQLHSTMT StatementHandle,
   SQLINTEGER TextLength)
 {
   MADB_Stmt* Stmt= (MADB_Stmt*)StatementHandle;
+  RESET_CANCELED(Stmt);
 
   MDBUG_C_ENTER(Stmt->Connection, "SQLPrepare");
 
@@ -1944,7 +2024,7 @@ SQLRETURN MA_SQLPrepare(SQLHSTMT StatementHandle,
   MDBUG_C_DUMP(Stmt->Connection, TextLength, d);
   try
   {
-    return Stmt->Prepare((char*)StatementText, TextLength, Stmt->Connection->Dsn->PrepareOnClient == '\0');
+    return Stmt->Prepare((char*)StatementText, TextLength, !Stmt->Options.PrepareOnClient);
   }
   catch (SQLException &e)
   {
@@ -1969,6 +2049,7 @@ SQLRETURN MA_SQLPrepareW(SQLHSTMT StatementHandle,
   SQLRETURN ret;
   BOOL ConversionError;
 
+  RESET_CANCELED(Stmt);
   MDBUG_C_ENTER(Stmt->Connection, "SQLPrepareW");
 
   try // Currently string transcoding cannot throw, but in future it most cetainly will
@@ -1986,7 +2067,7 @@ SQLRETURN MA_SQLPrepareW(SQLHSTMT StatementHandle,
     }
     else
     {
-      ret= Stmt->Prepare(StmtStr, (SQLINTEGER)StmtLength, Stmt->Connection->Dsn->PrepareOnClient == '\0');
+      ret= Stmt->Prepare(StmtStr, (SQLINTEGER)StmtLength, !Stmt->Options.PrepareOnClient);
     }
   }
   catch (SQLException &e)
@@ -2300,6 +2381,7 @@ SQLRETURN MA_SQLRowCount(SQLHSTMT StatementHandle,
   SQLLEN* RowCountPtr)
 {
   MADB_Stmt* Stmt= (MADB_Stmt*)StatementHandle;
+  RESET_CANCELED(Stmt);
   return Stmt->Methods->RowCount(Stmt, RowCountPtr);
 }
 /* }}} */
@@ -2341,6 +2423,7 @@ SQLRETURN MA_SQLSetCursorName(SQLHSTMT StatementHandle,
   SQLSMALLINT NameLength)
 {
   MADB_Stmt* Stmt= (MADB_Stmt*)StatementHandle;
+  RESET_CANCELED(Stmt);
   return Stmt->Methods->SetCursorName(Stmt, (char*)CursorName, NameLength);
 }
 /* }}} */
@@ -2355,6 +2438,7 @@ SQLRETURN MA_SQLSetCursorNameW(SQLHSTMT StatementHandle,
   SQLULEN Length;
   SQLRETURN ret;
 
+  RESET_CANCELED(Stmt);
   try
   {
     CpName= MADB_ConvertFromWChar(CursorName, NameLength, &Length, Stmt->Connection->ConnOrSrcCharset, NULL);
@@ -2364,6 +2448,10 @@ SQLRETURN MA_SQLSetCursorNameW(SQLHSTMT StatementHandle,
   {
     ret= MADB_SetError(&Stmt->Error, MADB_ERR_HY001, NULL, 0);
   }
+  catch (SQLException &e)
+  {
+    ret= MADB_FromException(Stmt->Error, e);
+  }
 
   MADB_FREE(CpName);
 
@@ -2432,6 +2520,8 @@ SQLRETURN MA_SQLSetStmtAttr(SQLHSTMT StatementHandle,
   if (!Stmt)
     return SQL_INVALID_HANDLE;
 
+  RESET_CANCELED(Stmt);
+
   MDBUG_C_ENTER(Stmt->Connection, "SQLSetStmtAttr");
   MDBUG_C_DUMP(Stmt->Connection, Attribute, d);
   MDBUG_C_DUMP(Stmt->Connection, ValuePtr, 0x);
@@ -2785,6 +2875,7 @@ SQLRETURN MA_SQLSetScrollOptions(SQLHSTMT hstmt,
   SQLUSMALLINT crowRowSet)
 {
   MADB_Stmt* Stmt= (MADB_Stmt*)hstmt;
+  RESET_CANCELED(Stmt);
   return MADB_DescSetField(Stmt->Ard, 0, SQL_DESC_ARRAY_SIZE, (SQLPOINTER)(SQLULEN)crowKeySet, SQL_IS_USMALLINT, 0);
 }
 /* }}} */
@@ -2798,6 +2889,7 @@ SQLRETURN MA_SQLParamOptions(
   MADB_Stmt* Stmt= (MADB_Stmt*)hstmt;
   SQLRETURN result;
 
+  RESET_CANCELED(Stmt);
   result= MADB_DescSetField(Stmt->Apd, 0, SQL_DESC_ARRAY_SIZE, (SQLPOINTER)crow, SQL_IS_UINTEGER, 0);
 
   if (SQL_SUCCEEDED(result))
